/*
 * Project: MoleCuilder
 * Description: creates and alters molecular systems
 * Copyright (C)  2013 Frederik Heber. All rights reserved.
 * Please see the COPYING file or "Copyright notice" in builder.cpp for details.
 * 
 *
 *   This file is part of MoleCuilder.
 *
 *    MoleCuilder is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    MoleCuilder is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with MoleCuilder.  If not, see <http://www.gnu.org/licenses/>. 
 */

/*
 * PairPotential_LennardJones.cpp
 *
 *  Created on: Jul 05, 2013
 *      Author: heber
 */


// include config.h
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "CodePatterns/MemDebug.hpp"

#include "PairPotential_LennardJones.hpp"

#include <boost/assign/list_of.hpp> // for 'map_list_of()'
#include <boost/bind.hpp>
#include <boost/lambda/lambda.hpp>
#include <cmath>
#include <string>

#include "CodePatterns/Assert.hpp"

#include "FunctionApproximation/Extractors.hpp"
#include "FunctionApproximation/TrainingData.hpp"
#include "Potentials/helpers.hpp"
#include "Potentials/ParticleTypeCheckers.hpp"

class Fragment;

// static definitions
const PairPotential_LennardJones::ParameterNames_t
PairPotential_LennardJones::ParameterNames =
      boost::assign::list_of<std::string>
      ("epsilon")
      ("sigma")
    ;
const std::string PairPotential_LennardJones::potential_token("lennardjones");

void PairPotential_LennardJones::setDefaultParameters()
{
  params[epsilon] = 1e-5;
  params[sigma] = 8.2;
}

PairPotential_LennardJones::PairPotential_LennardJones() :
    EmpiricalPotential(),
    params(parameters_t(MAXPARAMS, 0.))
{
  // have some decent defaults for parameter_derivative checking
  setDefaultParameters();
}

PairPotential_LennardJones::PairPotential_LennardJones(
    const ParticleTypes_t &_ParticleTypes
    ) :
  EmpiricalPotential(_ParticleTypes),
  params(parameters_t(MAXPARAMS, 0.))
{
  // have some decent defaults for parameter_derivative checking
  setDefaultParameters();
}

PairPotential_LennardJones::PairPotential_LennardJones(
    const ParticleTypes_t &_ParticleTypes,
    const double _epsilon,
    const double _sigma
    ) :
  EmpiricalPotential(_ParticleTypes),
  params(parameters_t(MAXPARAMS, 0.))
{
  params[epsilon] = _epsilon;
  params[sigma] = _sigma;
}

void PairPotential_LennardJones::setParameters(const parameters_t &_params)
{
  const size_t paramsDim = _params.size();
  ASSERT( paramsDim <= getParameterDimension(),
      "PairPotential_LennardJones::setParameters() - we need not more than "
      +toString(getParameterDimension())+" parameters.");
  for(size_t i=0;i<paramsDim;++i)
    params[i] = _params[i];

#ifndef NDEBUG
  parameters_t check_params(getParameters());
  check_params.resize(paramsDim); // truncate to same size
  ASSERT( check_params == _params,
      "PairPotential_LennardJones::setParameters() - failed, mismatch in to be set "
      +toString(_params)+" and set "+toString(check_params)+" params.");
#endif
}

PairPotential_LennardJones::results_t
PairPotential_LennardJones::operator()(
    const arguments_t &arguments
    ) const
{
  ASSERT( arguments.size() == 1,
      "PairPotential_LennardJones::operator() - requires one argument.");
  ASSERT( ParticleTypeChecker::checkArgumentsAgainstParticleTypes(
      arguments, getParticleTypes()),
      "PairPotential_LennardJones::operator() - types don't match with ones in arguments.");
  const double &r = arguments[0].distance;
  const double temp = Helpers::pow(params[sigma]/r, 6);
  const result_t result = 4.*params[epsilon] * (temp*temp - temp);
  return std::vector<result_t>(1, result);
}

PairPotential_LennardJones::derivative_components_t
PairPotential_LennardJones::derivative(
    const arguments_t &arguments
    ) const
{
  ASSERT( arguments.size() == 1,
      "PairPotential_LennardJones::operator() - requires no argument.");
  ASSERT( ParticleTypeChecker::checkArgumentsAgainstParticleTypes(
      arguments, getParticleTypes()),
      "PairPotential_LennardJones::operator() - types don't match with ones in arguments.");
  const double &r = arguments[0].distance;
  const double sigma6 = Helpers::pow(params[sigma], 6);
  const result_t result =
      4.*params[epsilon] * (
          sigma6*sigma6*(-12.) / Helpers::pow(r,13)
          - sigma6*(-6.) /Helpers::pow(r,7)
      );
  derivative_components_t results(1, result);
  return results;
}

PairPotential_LennardJones::results_t
PairPotential_LennardJones::parameter_derivative(
    const arguments_t &arguments,
    const size_t index
    ) const
{
  ASSERT( arguments.size() == 1,
      "PairPotential_LennardJones::parameter_derivative() - requires no argument.");
  ASSERT( ParticleTypeChecker::checkArgumentsAgainstParticleTypes(
      arguments, getParticleTypes()),
      "PairPotential_LennardJones::operator() - types don't match with ones in arguments.");
  const double &r = arguments[0].distance;
  switch (index) {
    case epsilon:
    {
      const double temp = Helpers::pow(params[sigma]/r, 6);
      const result_t result = 4. * (temp*temp - temp);
      return std::vector<result_t>(1, result);
      break;
    }
    case sigma:
    {
      const double r6 = Helpers::pow(r, 6);
      const result_t result =
          4.*params[epsilon] * (
              12. * Helpers::pow(params[sigma],11)/(r6*r6)
              - 6. * Helpers::pow(params[sigma],5)/r6
          );
      return std::vector<result_t>(1, result);
      break;
    }
    default:
      break;
  }
  return std::vector<result_t>(1, 0.);
}

FunctionModel::extractor_t
PairPotential_LennardJones::getSpecificExtractor() const
{
  Fragment::charges_t charges;
  charges.resize(getParticleTypes().size());
  std::transform(getParticleTypes().begin(), getParticleTypes().end(),
      charges.begin(), boost::lambda::_1);
  FunctionModel::extractor_t returnfunction =
      boost::bind(&Extractors::gatherDistancesFromFragment,
          boost::bind(&Fragment::getPositions, _1),
          boost::bind(&Fragment::getCharges, _1),
          charges,
          _2);
  return returnfunction;
}

FunctionModel::filter_t PairPotential_LennardJones::getSpecificFilter() const
{
  FunctionModel::filter_t returnfunction =
      boost::bind(&Extractors::filterArgumentsByParticleTypes,
          _1,
          getParticleTypes());
  return returnfunction;
}

void
PairPotential_LennardJones::setParametersToRandomInitialValues(
    const TrainingData &data)
{
  params[PairPotential_LennardJones::epsilon] = 1e-2*rand()/(double)RAND_MAX;
  params[PairPotential_LennardJones::sigma] = (3.+10.*rand()/(double)RAND_MAX);// 0.5;
}

